"""Re-add user agent to db

Revision ID: 244273a86d81
Revises: 54aecbd6187c
Create Date: 2020-06-30 15:03:18.260742

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '244273a86d81'
down_revision = '54aecbd6187c'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.add_column(sa.Column('user_agent', sa.String(length=255), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.drop_column('user_agent')

    # ### end Alembic commands ###
