"""Add double puppet base URL to puppet table

Revision ID: 3d4f3a252006
Revises: 3c59aa2a3af9
Create Date: 2020-09-18 20:06:14.910573

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3d4f3a252006'
down_revision = '3c59aa2a3af9'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('puppet', schema=None) as batch_op:
        batch_op.add_column(sa.Column('base_url', sa.Text(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('puppet', schema=None) as batch_op:
        batch_op.drop_column('base_url')
    # ### end Alembic commands ###
