# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from mautrix.types import ExtensibleEnum, SerializableEnum


class MessageUnsendability(ExtensibleEnum):
    DENY_FOR_NON_SENDER = "deny_for_non_sender"
    DENY_LOG_MESSAGE = "deny_log_message"
    DENY_TOMBSTONE_MESSAGE = "deny_tombstone_message"
    DENY_IF_PAGE_THREAD = "deny_if_page_thread"
    DENY_IF_CANNOT_REPLY_TO_VIEWER_THREAD = "deny_if_cannot_reply_to_viewer_thread"
    DENY_IF_MARKETPLACE_THREAD = "deny_if_marketplace_thread"
    CAN_UNSEND = "can_unsend"


class ThreadFolder(SerializableEnum):
    INBOX = "INBOX"
    PENDING = "PENDING"
    ARCHIVED = "ARCHIVED"
    OTHER = "OTHER"
